unit MainUnit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, XMLCfg, XMLConf, FileUtil, LResources, Forms, Controls,
  Graphics, Dialogs, StdCtrls, EditBtn, ExtCtrls;

type

  { TMapInfo }
  TMapInfo = record
    Name: string;
    Game: string;
  end;

  { TMain }

  TMain = class(TForm)
    ApplicationProperties1: TApplicationProperties;
    Bevel1: TBevel;
    Bevel2: TBevel;
    btRefresh: TButton;
    btRun: TButton;
    btExit: TButton;
    btQuakeInjector: TButton;
    cbHipnotic: TCheckBox;
    cbRogue: TCheckBox;
    cbMode: TComboBox;
    cbWindow: TCheckBox;
    cbBpp: TCheckBox;
    cbSkill: TComboBox;
    fnQuakeDir: TDirectoryEdit;
    edArgs: TEdit;
    edCmdLine: TEdit;
    fnEngine: TFileNameEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    lbGames: TListBox;
    lbMaps: TListBox;
    Cfg: TXMLConfig;
    procedure btExitClick(Sender: TObject);
    procedure btQuakeInjectorClick(Sender: TObject);
    procedure btRefreshClick(Sender: TObject);
    procedure btRunClick(Sender: TObject);
    procedure cbBppChange(Sender: TObject);
    procedure cbHipnoticChange(Sender: TObject);
    procedure cbModeChange(Sender: TObject);
    procedure cbModeExit(Sender: TObject);
    procedure cbRogueChange(Sender: TObject);
    procedure cbSkillChange(Sender: TObject);
    procedure cbWindowChange(Sender: TObject);
    procedure edArgsChange(Sender: TObject);
    procedure edArgsExit(Sender: TObject);
    procedure fnEngineChange(Sender: TObject);
    procedure fnQuakeDirAcceptDirectory(Sender: TObject; var Value: String);
    procedure fnQuakeDirExit(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure lbGamesSelectionChange(Sender: TObject; User: boolean);
    procedure lbMapsSelectionChange(Sender: TObject; User: boolean);
  private
    { private declarations }
  public
    { public declarations }
    Maps: array of TMapInfo;
    HasHipnotic, HasRogue, ExpansionsAutoSelected: Boolean;

    procedure AddMap(Game, MapName: string);

    procedure CheckExpansionsState;
    function CheckDirectory: Boolean;
    procedure ScanQuakeDir(ADir, APrefix: string);
    procedure ScanGameDir(ADir, APrefix: string);
    procedure UpdateGamesList;
    procedure UpdateMapsList(AGame: string);
    procedure UpdateCommandLine;

    procedure LoadConfiguration;
    procedure SaveConfiguration;
  end; 

var
  Main: TMain;

implementation

{$R *.lfm}

{ TMain }

procedure TMain.btExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMain.btQuakeInjectorClick(Sender: TObject);
begin
  try
    ChDir(fnQuakeDir.Text);
    ExecuteProcess('cmd', '/c start quakeinjector.jar');
  except
    MessageDlg('Error', 'Failed to launch Quake Injector', mtError, [mbOK], 0);
  end;
end;

procedure TMain.btRefreshClick(Sender: TObject);
begin
  if not CheckDirectory then Exit;
  ScanQuakeDir(fnQuakeDir.Directory, '');
end;

procedure TMain.btRunClick(Sender: TObject);
var
  Cmd, Arg: string;
begin
  if not CheckDirectory then Exit;
  if fnEngine.Text='' then begin;
    ShowMessage('Select an engine executable to use');
    fnEngine.SetFocus;
    Exit;
  end;
  if not FileExists(fnEngine.Text) then begin
    ShowMessage('The engine executable was not found. Please select a valid engine executable');
    fnEngine.SetFocus;
    Exit;
  end;
  SaveConfiguration;
  UpdateCommandLine;
  ChDir(fnQuakeDir.Text);
  Cmd:=Copy(edCmdLine.Text, 1, Pos(' ', edCmdLine.Text) - 1);
  Arg:=Copy(edCmdLine.Text, Pos(' ', edCmdLine.Text) + 1, Length(edCmdLine.Text));
  ExecuteProcess(Cmd, Arg, []);
end;

procedure TMain.cbBppChange(Sender: TObject);
begin
  UpdateCommandLine;
end;

procedure TMain.cbHipnoticChange(Sender: TObject);
begin
  if cbHipnotic.Checked and cbRogue.Checked then cbRogue.Checked:=False;
  UpdateCommandLine;
  SaveConfiguration;
end;

procedure TMain.cbModeChange(Sender: TObject);
begin
  UpdateCommandLine;
end;

procedure TMain.cbModeExit(Sender: TObject);
begin
  SaveConfiguration;
end;

procedure TMain.cbRogueChange(Sender: TObject);
begin
  if cbRogue.Checked and cbHipnotic.Checked then cbHipnotic.Checked:=False;
  UpdateCommandLine;
  SaveConfiguration;
end;

procedure TMain.cbSkillChange(Sender: TObject);
begin
  UpdateCommandLine;
end;

procedure TMain.cbWindowChange(Sender: TObject);
begin
  UpdateCommandLine;
  SaveConfiguration;
end;

procedure TMain.edArgsChange(Sender: TObject);
begin
  UpdateCommandLine;
end;

procedure TMain.edArgsExit(Sender: TObject);
begin
  SaveConfiguration;
end;

procedure TMain.fnEngineChange(Sender: TObject);
begin
  UpdateCommandLine;
end;

procedure TMain.fnQuakeDirAcceptDirectory(Sender: TObject; var Value: String);
begin
  fnQuakeDir.Text:=Value;
  btRefreshClick(Sender);
  SaveConfiguration;
end;

procedure TMain.fnQuakeDirExit(Sender: TObject);
begin
  if DirectoryExists(fnQuakeDir.Text) then begin
    if (fnEngine.Text='') and FileExists(fnQuakeDir.Text + DirectorySeparator + 'glquake.exe') then fnEngine.Text:=fnQuakeDir.Text + DirectorySeparator + 'glquake.exe';
    if lbGames.Items.Count=0 then btRefreshClick(Sender);
  end;
  UpdateCommandLine;
  SaveConfiguration;
end;

procedure TMain.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  SaveConfiguration;
end;

procedure TMain.FormCreate(Sender: TObject);
begin
  ForceDirectories(GetAppConfigDir(False));
  Cfg.Filename:=GetAppConfigDir(False) + 'miniql.cfg';
  LoadConfiguration;
end;

procedure TMain.lbGamesSelectionChange(Sender: TObject; User: boolean);
begin
  if lbGames.ItemIndex <> -1 then UpdateMapsList(lbGames.Items[lbGames.ItemIndex]);
end;

procedure TMain.lbMapsSelectionChange(Sender: TObject; User: boolean);
begin
  UpdateCommandLine;
end;

procedure TMain.AddMap(Game, MapName: string);
var
  MapInfo: TMapInfo;
begin
  if LowerCase(Game)='id1' then Game:='(id1)'
  else if LowerCase(Game)='rogue' then Game:='(rogue)'
  else if LowerCase(Game)='hipnotic' then Game:='(hipnotic)';
  if LowerCase(MapName)='start' then MapName:='(start)';
  MapInfo.Game:=Game;
  MapInfo.Name:=MapName;
  SetLength(Maps, Length(Maps) + 1);
  Maps[Length(Maps) - 1]:=MapInfo;
end;

procedure TMain.CheckExpansionsState;
begin
  cbHipnotic.Enabled:=lbGames.Items.IndexOf('(hipnotic)') > -1;
  cbRogue.Enabled:=lbGames.Items.IndexOf('(rogue)') > -1;
end;

function TMain.CheckDirectory: Boolean;
begin
  if fnQuakeDir.Directory='' then begin
    ShowMessage('Select the directory where Quake is located');
    fnQuakeDir.SetFocus;
    Exit(False);
  end;
  if not DirectoryExists(fnQuakeDir.Directory) then begin
    ShowMessage('The directory "' + fnQuakeDir.Directory + '" does not exist. Please give the correct directory path');
    fnQuakeDir.SetFocus;
    Exit(False);
  end;
  Result:=True;
end;

procedure TMain.ScanQuakeDir(ADir, APrefix: string);
var
  fi: TSearchRec;
  r: LongInt;
begin
  if APrefix='' then begin
    SetLength(Maps, 0);
    HasHipnotic:=False;
    HasRogue:=False;
    btQuakeInjector.Visible:=FileExists(ADir + DirectorySeparator + 'quakeinjector.jar');
  end;
  r:=FindFirst(ADir + DirectorySeparator + '*.*', faAnyFile - faVolumeId, fi);
  if r=0 then repeat
    if (fi.Attr and faDirectory)=faDirectory then begin
      if (fi.Name <> '.') and (fi.Name <> '..') then
        ScanGameDir(ADir + DirectorySeparator + fi.Name, APrefix);
    end;
    r:=FindNext(fi);
  until r <> 0;
  FindClose(fi);
  if APrefix='' then UpdateGamesList;
end;

procedure TMain.ScanGameDir(ADir, APrefix: string);
var
  Game: string;

  procedure ScanPAKFile(APakFile: string);
  type
    TEntry = packed record
      Name: array [0..55] of Char;
      Pos: Cardinal;
      Len: Cardinal;
    end;
  var
    i, DirPos, DirLen, Files: Cardinal;
    F: TFileStream;
    Head: array [0..3] of Char;
    Entry: TEntry;
    MapName: string;
  begin
    F:=TFileStream.Create(APakFile, fmOpenRead);
    if F.Read(Head, 4) < 4 then begin
      F.Free;
      Exit;
    end;
    if Head <> 'PACK' then begin
      F.Free;
      Exit;
    end;
    DirPos:=F.ReadDWord;
    DirLen:=F.ReadDWord;
    Files:=DirLen div SizeOf(TEntry);
    F.Position:=DirPos;
    for i:=1 to Files do begin
      F.Read(Entry, SizeOf(TEntry));
      if (Copy(Entry.Name, 1, 5)='maps/') and (Pos('.bsp', Entry.Name) > 0) then begin
        MapName:=ExtractFileNameWithoutExt(Entry.Name);
        AddMap(Game, Copy(MapName, 6, Length(MapName)));
      end;
    end;
    F.Free;
  end;

  procedure ScanPAKFiles;
  var
    i: Integer;
    fn: string;
  begin
    for i:=0 to 255 do begin
      fn:=ADir + DirectorySeparator + 'pak' + IntToStr(i) + '.pak';
      if FileExists(fn) then ScanPAKFile(fn) else Break;
    end;
  end;

  procedure ScanMapsDir;
  var
    fi: TSearchRec;
    r: LongInt;
  begin
    r:=FindFirst(ADir + DirectorySeparator + 'maps' + DirectorySeparator + '*.bsp', faAnyFile - faVolumeId, fi);
    if r=0 then repeat
      AddMap(Game, ExtractFileNameWithoutExt(fi.Name));
      r:=FindNext(fi);
    until r <> 0;
    FindClose(fi);
  end;

begin
  Game:=APrefix + ExtractFileName(ADir);
  if LowerCase(Game)='hipnotic' then HasHipnotic:=True;
  if LowerCase(Game)='rogue' then HasRogue:=True;
  if FileExists(ADir + DirectorySeparator + 'pak0.pak') then ScanPAKFiles;
  if DirectoryExists(ADir + DirectorySeparator + 'maps') then ScanMapsDir;
  ScanQuakeDir(ADir, Copy(ADir, Length(fnQuakeDir.Directory) + 2, MaxInt) + DirectorySeparator);
end;

procedure TMain.UpdateGamesList;
var
  i: Integer;
begin
  lbGames.Clear;
  lbGames.Items.Add('(all)');
  lbGames.Items.Add('(id1)');
  if HasHipnotic then lbGames.Items.Add('(hipnotic)');
  if HasRogue then lbGames.Items.Add('(rogue)');
  for i:=0 to Length(Maps)-1 do begin
    if lbGames.Items.IndexOf(Maps[i].Game)=-1 then lbGames.Items.Add(Maps[i].Game);
  end;
  if lbGames.Items.IndexOf('(id1)') <> -1 then lbGames.ItemIndex:=lbGames.Items.IndexOf('(id1)')
end;

procedure TMain.UpdateMapsList(AGame: string);
var
  i: Integer;
begin
  CheckExpansionsState;
  if ExpansionsAutoSelected then begin
    cbHipnotic.Checked:=False;
    cbRogue.Checked:=False;
    ExpansionsAutoSelected:=False;
  end;
  lbMaps.Clear;
  for i:=0 to Length(Maps)-1 do begin
    if AGame='(all)' then begin
      if Maps[i].Name='(start)' then Continue;
      lbMaps.Items.Add(Maps[i].Name + ' of ' + Maps[i].Game);
    end else if Maps[i].Game=AGame then begin
      lbMaps.Items.Add(Maps[i].Name);
    end;
  end;
  if lbMaps.Items.IndexOf('(start)') <> -1 then lbMaps.ItemIndex:=lbMaps.Items.IndexOf('(start)')
  else if lbMaps.Items.Count=1 then lbMaps.ItemIndex:=0
  else if lbMaps.Items.IndexOf(AGame) <> -1 then lbMaps.ItemIndex:=lbMaps.Items.IndexOf(AGame);
  if AGame='(id1)' then begin
    cbHipnotic.Checked:=False;
    cbRogue.Checked:=False;
    cbHipnotic.Enabled:=False;
    cbRogue.Enabled:=False;
    ExpansionsAutoSelected:=True;
  end else if AGame='(hipnotic)' then begin
    cbHipnotic.Checked:=True;
    cbRogue.Checked:=False;
    cbHipnotic.Enabled:=False;
    cbRogue.Enabled:=False;
    ExpansionsAutoSelected:=True;
  end else if AGame='(rogue)' then begin
    cbHipnotic.Checked:=False;
    cbRogue.Checked:=True;
    cbHipnotic.Enabled:=False;
    cbRogue.Enabled:=False;
    ExpansionsAutoSelected:=True;
  end;
  UpdateCommandLine;
end;

procedure TMain.UpdateCommandLine;
var
  Cmd, Game, MapName: string;
begin
  if (lbGames.ItemIndex=-1) or (lbMaps.ItemIndex=-1) then Exit;
  Cmd:=ExtractFileName(fnEngine.Text);
  if cbHipnotic.Checked then Cmd:=Cmd + ' -hipnotic';
  if cbRogue.Checked then Cmd:=Cmd + ' -rogue';
  Game:=lbGames.Items[lbGames.ItemIndex];
  MapName:=lbMaps.Items[lbMaps.ItemIndex];
  if (Game <> '') and (Game[1] <> '(') then Cmd:=Cmd + ' -game ' + Game;
  if (cbMode.Text <> '') and (cbMode.Text <> '(default)') then begin
    if cbMode.Text='(desktop)' then
      Cmd:=Cmd + ' -width ' + IntToStr(Screen.MonitorFromWindow(Handle).Width) + ' -height ' + IntToStr(Screen.MonitorFromWindow(Handle).Height)
    else
      Cmd:=Cmd + ' -width ' + cbMode.Text;
  end;
  if cbBpp.Checked then Cmd:=Cmd + ' -bpp 32';
  if cbWindow.Checked then Cmd:=Cmd + ' -window';
  if edArgs.Text <> '' then Cmd:=Cmd + ' ' + edArgs.Text;
  if cbSkill.ItemIndex <> 0 then Cmd:=Cmd + ' +skill ' + IntToStr(cbSkill.ItemIndex - 1);
  if MapName <> '' then begin
    if MapName='(start)' then Cmd:=Cmd + ' +map start'
    else Cmd:=Cmd + ' +map ' + MapName;
  end;
  edCmdLine.Text:=Trim(Cmd);
end;

procedure TMain.LoadConfiguration;
begin
  fnQuakeDir.Text:=Cfg.GetValue('Game/QuakeDirectory', '');
  fnEngine.Text:=Cfg.GetValue('Game/EngineExecutable', '');
  cbHipnotic.Checked:=Cfg.GetValue('Expansions/HipnoticChecked', False);
  cbRogue.Checked:=Cfg.GetValue('Expansions/RogueChecked', False);
  cbBpp.Checked:=Cfg.GetValue('Options/Bpp32Checked', True);
  cbWindow.Checked:=Cfg.GetValue('Options/WindowChecked', False);
  cbMode.Text:=Cfg.GetValue('Options/Mode', '(desktop)');
  cbSkill.ItemIndex:=Cfg.GetValue('Options/Skill', 2);
  edArgs.Text:=Cfg.GetValue('Misc/Arguments', '');

  if fnQuakeDir.Text <> '' then fnQuakeDirExit(Cfg);
end;

procedure TMain.SaveConfiguration;
begin
  Cfg.SetValue('Game/QuakeDirectory', fnQuakeDir.Text);
  Cfg.SetValue('Game/EngineExecutable', fnEngine.Text);
  Cfg.SetValue('Expansions/HipnoticChecked', cbHipnotic.Checked);
  Cfg.SetValue('Expansions/RogueChecked', cbRogue.Checked);
  Cfg.SetValue('Options/Bpp32Checked', cbBpp.Checked);
  Cfg.SetValue('Options/WindowChecked', cbWindow.Checked);
  Cfg.SetValue('Options/Mode', cbMode.Text);
  Cfg.SetValue('Options/Skill', cbSkill.ItemIndex);
  Cfg.SetValue('Misc/Arguments', edArgs.Text);
  Cfg.Flush;
end;

end.

